# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from mock import MagicMock

from powerfulseal.policy.action_kubectl import ActionKubectl
from powerfulseal.policy.action_probe_http import ActionProbeHTTP
from powerfulseal.policy.action_nodes import ActionNodes
from powerfulseal.policy.action_pods import ActionPods
from powerfulseal.policy.action_nodes_pods import ActionNodesPods
from powerfulseal.execute import SSHExecutor


# Common fixtures
class Dummy():
    pass

def make_dummy_object():
    return Dummy()

@pytest.fixture
def dummy_object():
    return make_dummy_object()


# ActionNodesPods fixtures
@pytest.fixture
def noop_scenario():
    scenario = ActionNodesPods(
        name="test scenario",
        schema={}
    )
    scenario.match = MagicMock(return_value=[])
    scenario.filter = MagicMock(return_value=[])
    scenario.act = MagicMock()
    return scenario


@pytest.fixture
def no_filtered_items_scenario():
    scenario = ActionNodesPods(
        name="test scenario",
        schema={}
    )
    matched_items = [ make_dummy_object() ]
    scenario.match = MagicMock(return_value=matched_items)
    scenario.filter = MagicMock(return_value=[])
    scenario.act = MagicMock()
    return scenario


# Pod ActionNodesPods Fixtures
EXAMPLE_POD_SCHEMA = {
    "matches": [
        {
            "property": {
                "name": "attr",
                "value": "a.*"
            }
        },
    ]
}


@pytest.fixture
def pod_scenario():
    inventory = MagicMock()
    k8s_inventory = MagicMock()
    executor = SSHExecutor()
    return ActionPods(
        name="test scenario",
        schema=EXAMPLE_POD_SCHEMA,
        inventory=inventory,
        k8s_inventory=k8s_inventory,
        executor=executor,
    )


# Node ActionNodesPods Fixtures
EXAMPLE_NODE_SCHEMA = {
    "matches": [
        {
            "property": {
                "name": "attr",
                "value": "a.*"
            }
        },
    ]
}


@pytest.fixture
def node_scenario():
    inventory = MagicMock()
    driver = MagicMock()
    executor = MagicMock()
    return ActionNodes(
        name="test scenario",
        schema=EXAMPLE_NODE_SCHEMA,
        inventory=inventory,
        driver=driver,
        executor=executor,
    )


@pytest.fixture
def action_kubectl():
    logger = MagicMock()
    return ActionKubectl(
        name="test kubectl action",
        schema=dict(
            action="apply",
            payload="---\n"
        ),
        logger=logger,
    )


@pytest.fixture
def action_probe_http():
    logger = MagicMock()
    k8s_inventory = MagicMock()
    return ActionProbeHTTP(
        name="test probe http action",
        schema=dict(
            target=dict(
                url="http://example.com"
            )
        ),
        k8s_inventory=k8s_inventory,
        logger=logger,
    )



import pytest
from mock import patch, MagicMock
from powerfulseal.clouddrivers import aws_driver
from powerfulseal.node import Node

PRIVATE_IPS = ['198.168.1.1', '198.168.2.1']
PUBLIC_IPS = ['31.31.31.31', '41.41.41.41']
INVALID_IP = '198.168.3.1'

class EC2instance():
    def __init__(self, id, private_ip_address, public_ip_address, zone, state):
        self.private_ip_address = private_ip_address
        self.id = id
        self.public_ip_address = public_ip_address
        self.placement = {}
        self.placement = {'AvailabilityZone': zone}
        self.state = {'Name': state}

@pytest.fixture
def ec2_instances():
    return [
        EC2instance(id="i-123456789", private_ip_address="198.168.1.1", zone="us-east-1a", public_ip_address="31.31.31.31", state="Running"),
        EC2instance(id="i-987654321", private_ip_address="198.168.2.1", zone="us-east-1b", public_ip_address="41.41.41.41", state="Running"),
]

@patch('powerfulseal.clouddrivers.aws_driver.create_connection_from_config')
def test_get_by_ip_private_ip_nodes(create_connection_from_config, ec2_instances):
    driver = aws_driver.AWSDriver()
    driver.remote_servers = ec2_instances
    node_index = 0
    nodes = driver.get_by_ip(PRIVATE_IPS[node_index])
    assert ec2_instances[node_index].id is nodes.id
    assert ec2_instances[node_index].placement['AvailabilityZone'] is nodes.az
    assert ec2_instances[node_index].private_ip_address == nodes.ip

@patch('powerfulseal.clouddrivers.aws_driver.create_connection_from_config')
def test_get_by_ip_public_ip_nodes(create_connection_from_config, ec2_instances):
    driver = aws_driver.AWSDriver()
    driver.remote_servers = ec2_instances
    node_index = 1
    nodes = driver.get_by_ip(PUBLIC_IPS[node_index])
    assert ec2_instances[node_index].id is nodes.id
    assert ec2_instances[node_index].placement['AvailabilityZone'] is nodes.az
    assert ec2_instances[node_index].public_ip_address == nodes.extIp

@patch('powerfulseal.clouddrivers.aws_driver.create_connection_from_config')
def test_get_by_ip_no_nodes(create_connection_from_config, ec2_instances):
    driver = aws_driver.AWSDriver()
    driver.remote_servers = ec2_instances
    nodes = driver.get_by_ip(INVALID_IP)
    assert nodes is None




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import configparser

ConfigParser = configparser.ConfigParser


def read_inventory_file_to_dict(inventory_filename):
    """ Reads inventory file in ini format (for example in Ansible format),
        returns a dictionary, where keys correspond to lower case sections,
        and resolved first level-inclusions
    """
    config = configparser.ConfigParser(allow_no_value=True)
    config.read(inventory_filename)
    # extract the IPs into lower-case groups per section
    _groups = {
        key.lower(): [ ip.lower().split(" ")[0] for (ip, group) in config.items(key)]
        for key in config.sections()
    }
    # resolve one level groups hierarchy
    groups = {}
    for group, ips in _groups.items():
        ips_set = set()
        for ip in ips:
            subgroup = _groups.get(ip)
            if subgroup is not None:
                for sub in subgroup:
                    ips_set.add(sub)
            else:
                ips_set.add(ip)
        groups[group] = list(ips_set)
    return groups



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os.path import expanduser
from powerfulseal.cli.__main__ import parse_args

HOME = expanduser("~")


def test_validate_policy_file_does_not_require_other_arguments():
    parser = parse_args([
        'validate',
        '--policy-file', 'test/config.yml'
    ])
    assert parser.mode == "validate"
    assert parser.policy_file == 'test/config.yml'
    
def test_kubeconfig_default():
    parser = parse_args([
        'interactive',
        '--inventory-kubernetes',
        '--no-cloud'
    ])
    assert parser.kubeconfig is None


def test_interactive_mode_integration():
    parser = parse_args([
        'interactive',
        '--inventory-kubernetes',
        '--kubeconfig', '~/.kube/config',
        '--no-cloud'
    ])
    assert parser.mode == "interactive"
    assert parser.inventory_kubernetes
    assert parser.kubeconfig == HOME + '/.kube/config'
    assert parser.no_cloud


def test_autonomous_mode_integration():
    parser = parse_args([
        'autonomous',
        '--inventory-kubernetes',
        '--kubeconfig', '~/config',
        '--no-cloud',
        '--policy-file', '~/policy.yml'
    ])
    assert parser.mode == "autonomous"
    assert parser.inventory_kubernetes
    assert parser.kubeconfig == HOME + '/config'
    assert parser.no_cloud
    assert parser.policy_file == '~/policy.yml'



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
from mock import patch, MagicMock

import powerfulseal
from powerfulseal.cli.__main__ import parse_args

HOME = os.path.expanduser("~")


def test_flag_takes_precedence(monkeypatch, tmpdir):
    p1 = tmpdir.join("kubeconfig1")
    p1.write("{}")
    p2 = tmpdir.join("kubeconfig2")
    p2.write("{}")
    monkeypatch.setenv('KUBECONFIG', str(p1.realpath()))
    with patch("powerfulseal.cli.__main__.KUBECONFIG_DEFAULT_PATH", str(p2.realpath())):
        kube_config = powerfulseal.cli.__main__.parse_kubeconfig(parse_args([
            'interactive',
            '--kubeconfig', '~/LOL',
            '--inventory-kubernetes',
            '--no-cloud'
        ]))
        assert kube_config == HOME + "/LOL"


def test_env_var_takes_2nd_precedence(monkeypatch, tmpdir):
    p1 = tmpdir.join("kubeconfig1")
    p1.write("{}")
    p2 = tmpdir.join("kubeconfig2")
    p2.write("{}")
    monkeypatch.setenv('KUBECONFIG', str(p1.realpath()))
    with patch("powerfulseal.cli.__main__.KUBECONFIG_DEFAULT_PATH", str(p2.realpath())):
        kube_config = powerfulseal.cli.__main__.parse_kubeconfig(parse_args([
            'interactive',
            '--inventory-kubernetes',
            '--no-cloud'
        ]))
        assert kube_config == str(p1.realpath())


def test_home_kubeconfig_takes_3rd_precedence(monkeypatch, tmpdir):
    p2 = tmpdir.join("kubeconfig2")
    p2.write("{}")
    monkeypatch.delenv('KUBECONFIG', raising=False)
    with patch("powerfulseal.cli.__main__.KUBECONFIG_DEFAULT_PATH", str(p2.realpath())):
        kube_config = powerfulseal.cli.__main__.parse_kubeconfig(parse_args([
            'interactive',
            '--inventory-kubernetes',
            '--no-cloud'
        ]))
        assert kube_config == str(p2.realpath())


def test_in_cluster_takes_4th_precedence(monkeypatch, tmpdir):
    p2 = tmpdir.join("kubeconfig2")
    p2.write("{}")
    monkeypatch.delenv('KUBECONFIG', raising=False)
    with patch("powerfulseal.cli.__main__.KUBECONFIG_DEFAULT_PATH", "/probably/doesnt/exist"):
        kube_config = powerfulseal.cli.__main__.parse_kubeconfig(parse_args([
            'interactive',
            '--inventory-kubernetes',
            '--no-cloud'
        ]))
        assert kube_config == None


# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from powerfulseal import makeLogger

import requests

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

POD_METRICS_PATH = "/apis/metrics.k8s.io/v1beta1/pods"

# Value is a fraction of a single CPU core (e.g., 1n = 1 / 1000000000 => 0.000000001)
CPU_UNITS = {'n': 1000000000}
# Value is a multiple of a byte (e.g., 1Ki => 1000)
MEMORY_UNITS = {'Ki': 1000, 'Mi': 1000000, 'Gi': 1000000000, 'Ti': 1000000000000, 'Pi': 1000000000000000}


class MetricsServerClient:
    """
    Retrieves CPU and memory metrics from metrics-server
    """

    def __init__(self, base_path, logger=None):
        self.base_path = base_path or ''
        self.logger = logger or makeLogger(__name__)

    def get_pod_metrics(self):
        response = requests.get(self.base_path + POD_METRICS_PATH)
        if response.status_code != 200:
            self.logger.error("Could not retrieve pod metrics. Exiting.")
            exit()

        try:
            data = response.json()
        except ValueError:
            self.logger.error("Unable to retrieve response data. Exiting.")
            exit()

        # Key: map of namespaces; value: map of pods
        metrics = {}
        for item in data['items']:
            namespace = item['metadata']['namespace']
            name = item['metadata']['name']
            cpu = sum(map(lambda c: self.parse_cpu_string(c['usage']['cpu']), item['containers']))
            memory = sum(map(lambda c: self.parse_memory_string(c['usage']['memory']), item['containers']))

            if namespace not in metrics:
                metrics[namespace] = {}

            metrics[namespace][name] = {
                'cpu': cpu,
                'memory': memory
            }

        return metrics

    def parse_cpu_string(self, cpu):
        """
        Returns CPU in number of cores (e.g., "120n" => 0.0000012)
        """
        if len(cpu) < 2 or is_numeric(cpu[-1]):
            return float(cpu)

        return float(cpu[:-1]) / CPU_UNITS[cpu[-1:]]

    def parse_memory_string(self, memory):
        """
        Returns memory in number of bytes (e.g., "228Ki" => 228000)
        """
        if len(memory) < 2:
            return int(memory)

        is_last_numeric = is_numeric(memory[-1])
        is_penultimate_numeric = is_numeric(memory[-2])
        if is_penultimate_numeric and is_last_numeric:
            # Memory is already in bytes
            return int(memory)
        elif not is_penultimate_numeric and is_last_numeric:
            # Impossible case of alphabetical followed by numeric
            raise ValueError("Memory is malformed")
        elif is_penultimate_numeric and not is_last_numeric:
            # Unimplemented case of single-character suffixes (it is unknown
            # whether metric-server implements this). Currently, memory is handled
            # with power-of-two suffixes (e.g., Ki, Mi, etc.)
            raise NotImplementedError("Single character units are not implemented")
        else:
            return int(memory[:-2]) * MEMORY_UNITS[memory[-2:]]


def is_numeric(c):
    return '0' <= c <= '9'




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import re

from powerfulseal import makeLogger
from powerfulseal.metriccollectors.collector import POD_SOURCE
from .action_nodes_pods import ActionNodesPods

class StartHostAction():
    """ A little helper class to start hosts in cleanup """
    def __init__(self, driver, host, logger=None):
        self.driver = driver
        self.host = host
        self.logger = logger or makeLogger(__name__)

    def execute(self):
        try:
            self.driver.start(self.host)
            self.logger.info("Restarted node %s", self.host)
        except:
            self.logger.exception("Exception restarting node %s", self.host)
            return False
        return True


class ActionPods(ActionNodesPods):
    """ Pod scenario handler.

        Adds matching for k8s-specific things and pod-specific actions
    """

    def __init__(self, name, schema, inventory, k8s_inventory, executor,
                 logger=None, metric_collector=None):
        ActionNodesPods.__init__(self, name, schema, logger=logger, metric_collector=metric_collector)
        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.executor = executor
        self.action_mapping = {
            "wait": self.action_wait,
            "kill": self.action_kill,
            "checkPodCount": self.action_check_pod_count,
            "checkPodState": self.action_check_pod_state,
            "stopHost": self.action_stop_host,
        }

    def match(self):
        """ Makes a union of all the pods matching any of the policy criteria.
        """
        mapping = {
            "namespace": self.match_namespace,
            "deployment": self.match_deployment,
            "labels": self.match_labels,
        }
        selected = set()
        criteria = self.schema.get("matches", [])
        self.logger.debug("Criteria %r ",criteria)
        for criterion in criteria:
            for action_name, action_method in mapping.items():
                if action_name in criterion:
                    self.logger.info("Matching %r %r", action_name, criterion)
                    params = criterion.get(action_name)
                    for pod in action_method(params):
                        self.logger.debug("Matching %r", pod)
                        selected.add(pod)
        if len(selected) == 0:
            self.metric_collector.add_matched_to_empty_set_metric(POD_SOURCE)
        return list(selected)

    def match_namespace(self, param):
        """ Matches pods for a namespace
        """
        namespace = param
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
        )
        self.logger.info("Matched %d pods in namespace %s", len(pods), namespace)
        return pods

    def match_deployment(self, params):
        """ Matches pods for a deployment
        """
        namespace = params.get("namespace")
        deployment_name = params.get("name")
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
            deployment_name=deployment_name,
        )
        self.logger.info("Matched %d pods for deploy %s in namespace %s",
                 len(pods), deployment_name, namespace
        )
        return pods

    def match_labels(self, params):
        """ Matches pods for a selector
        """
        namespace = params.get("namespace")
        selector = params.get("selector")
        pods = self.k8s_inventory.find_pods(
            namespace=namespace,
            selector=selector,
        )
        self.logger.info("Matched %d pods for selector %s in namespace %s",
                 len(pods), selector, namespace
        )
        return pods

    def action_kill(self, pods, params):
        """ Kills a pod by executing a docker kill on one of the containers or pod delete
        """
        probability = params.get("probability", 1)
        force = params.get("force", True)
        signal = "SIGKILL" if force else "SIGTERM"
        success = True
        for pod in pods:
            if probability < random.random():
                self.logger.info("Pod got lucky - not killing")
            else:
                ret_val = True
                try:
                    ret_val = self.executor.kill_pod(pod, self.inventory, signal)
                except:
                    self.logger.exception("Exception while killing pod")
                    ret_val = False
                # update the metrics
                if ret_val:
                    self.metric_collector.add_pod_killed_metric(pod)
                    self.logger.info("Pod killed: %s", pod)
                else:
                    self.metric_collector.add_pod_kill_failed_metric(pod)
                    self.logger.error("Pod NOT killed: %s", pod)
                    success = False
        return success

    def action_check_pod_count(self, pods, params):
        """
            Checks that the count of pods equals the desired count.
        """
        count = int(params.get("count"))
        if len(pods) != count:
            self.logger.error("Expected %d pods, got %d", count, len(pods))
            return False
        return True

    def action_check_pod_state(self, pods, params):
        """
            Checks that all the pods are in desired state.
        """
        state = params.get("state")
        success = True
        for pod in pods:
            if not self.match_property(
                candidate=pod,
                criterion=dict(
                    name="state",
                    value=state,
                )
            ):
                self.logger.error("Expected pod in state '%s', got '%s' (%r)", state, pod.state, pod)
                success = False
        return success

    def action_stop_host(self, pods, params):
        """ Action to stop a node.
        """
        self.inventory.sync()
        host_ips = list(set([p.host_ip for p in pods]))
        for host_ip in host_ips:
            host = self.inventory.get_node_by_ip(host_ip)
            if host is None:
                self.logger.warning("Couldn't find host with IP %r", host_ip)
                return False
            self.logger.info("Action stop on host %r (pods %s)", host, pods)
            try:
                self.inventory.driver.stop(host)
                if params.get("autoRestart", True):
                    self.cleanup_actions.append(
                        StartHostAction(driver=self.inventory.driver, host=host, logger=self.logger)
                    )
                self.metric_collector.add_node_stopped_metric(host)
            except:
                self.metric_collector.add_node_stop_failed_metric(host)
                self.logger.exception("Error stopping the machine")
                return False
        return True




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmd
import shlex
from termcolor import colored, cprint
import sys

try:
    import readline
    readline.set_completer_delims(" ,")
except:
    pass

from ..execute import (
    SSHExecutor,
)

DEFAULT_COLOR_KEYWORDS = {
    "UP": "green",
    "DOWN": "red",
    "node": "blue",
    "pod": "blue",
    "host_ip": "yellow",
    "ip": "yellow",
    "extIp": "yellow",
    "Running": "green",
    "Succeeded": "green",
    "Pending": "grey",
    "Failed": "red",
    "Unknown": "red",
    "CrashLoopBackOff": "red",
    "error": "red",
    "Error": "red",
    "namespace": "blue"
}
# couple of helpers
def colour_output(output, extras=None):
    if extras is not None:
        pattern = extras.copy()
        pattern.update(DEFAULT_COLOR_KEYWORDS)
    else:
        pattern = DEFAULT_COLOR_KEYWORDS
    for word, colour in pattern.items():
        output = output.replace(
            word+'=', colored(word, colour)+'=').replace('='+word, '='+colored(word, colour))
    return output

def filter_text_insensitive(collection, item=None):
    return [
        element
        for element in collection
        if item is None or element.lower().startswith(item.lower())
    ]

class Command():
    def __init__(self, line):
        self.line = line
        self.args = shlex.split(line)
        self.finished = line and (line[-1] == " ")

    def get(self, index, default=None):
        if index < 0 or index >= len(self.args):
            return default
        return self.args[index]

    def __len__(self):
        return len(self.args)


class PSCmd(cmd.Cmd):
    """
        PowerfulSeal cli base class.
    """

    def __init__(self, inventory, driver, executor, k8s_inventory):
        cmd.Cmd.__init__(self)
        self.inventory = inventory
        self.driver = driver
        self.prompt = "(seal) $ "
        self.executor = executor
        self.k8s_inventory = k8s_inventory

    def completedefault(self, text, line, begidx, endidx):
        suggestions = []
        text_lower = text.lower()
        try:
            # try to match AZ
            for az in self.inventory.get_azs():
                if text_lower in az.lower():
                    suggestions.append(str(az))
            # try to match group name
            for group in self.inventory.get_groups():
                if text_lower in group.lower():
                    suggestions.append(str(group))
            # try to match name, id, state, ip
            for node in self.inventory.find_nodes():
                for attr in ["ip", "id", "state", "name", "no"]:
                    val = str(getattr(node, attr))
                    if text_lower in val.lower():
                        suggestions.append(val)
            # extras
            for extra in ("all",):
                if extra.lower().startswith(text.lower()):
                    suggestions.append(extra)
        except Exception as e:
            # sadly, if an exception is thrown here, it's just
            # silently swallowed...
            print(e)

        return suggestions

    def find_nodes(self, query):
        nodes = list(self.inventory.find_nodes(query))
        if not nodes:
            print(colored("No matching hosts", "yellow"))
        return nodes

    ###########################################################################
    # CLI control
    ###########################################################################
    def do_exit(self, line):
        "Exit CLI"
        raise GeneratorExit

    ###########################################################################
    # NODE (MACHINE) RELATED FUNCTIONALITY
    ###########################################################################

    def do_nodes(self, line):
        """
        Prints the nodes about a particular node

        Syntax:
            nodes [<subset>]
        """
        cmd = Command(line)
        query = cmd.get(0)
        if query:
            extras = {query: "blue"}
        else:
            extras = {}
        for node in self.find_nodes(query):
            print(colour_output(str(node), extras))

    def do_sync(self, line):
        """
        Synchronises the state with the remote API
        """
        self.inventory.sync()
        self.do_nodes("")

    def do_zones(self, line):
        """
        Prints availability zones
        """
        for az in self.inventory.get_azs():
            print(az)

    def do_groups(self, line):
        """
        Prints the groups
        """
        for group in self.inventory.get_groups():
            print(group)

    def do_start(self, line):
        """
        Brings up a subset of machines
        """
        cmd = Command(line)
        for node in self.find_nodes(cmd.get(0)):
            print("Starting %s" % (node))
            try:
                self.driver.start(node)
            except Exception as e:
                print(e)

    def do_stop(self, line):
        """
        Brings down a subset of machines
        """
        cmd = Command(line)
        for node in self.find_nodes(cmd.get(0)):
            print("Stopping %s" % (node))
            try:
                self.driver.stop(node)
            except Exception as e:
                print(e)

    def do_delete(self, line):
        """
        PERMANENTLY DELETES A NODE

        Syntax:
            delete <subset>
        """
        deleted = 0
        cmd = Command(line)
        if cmd.get(0) is None:
            return print("Can't delete all machines at once. It's for your own good")
        for node in self.find_nodes(cmd.get(0)):
            print("About to PERMANENTLY DELETE THIS NODE: \n{node}".format(
                node=colour_output(str(node))
            ))
            answer = None
            while answer not in ("yes", "no"):
                sys.stdout.write("Proceed ? (yes|no): ")
                answer = input()
            if answer == "yes":
                print("Deleting")
                self.driver.delete(node)
                deleted += 1
            else:
                print("Skipping")
        print("Deleted {num} nodes".format(num=deleted))


    ###########################################################################
    # EXECUTION ON SSH RELATED FUNCTIONALITY
    ###########################################################################

    def execute(self, command, nodes):
        """
        Executes a line in shell on specified boxes
        """
        try:
            for node in nodes:
                for key, value in self.executor.execute(
                    command, nodes=[node]
                ).items():
                    if value["ret_code"] > 0:
                        print(colored("-" * 80, "red"))
                        print(colored(key, "red"))
                        print(colored(value.get("stdout", ""), "red"))
                        print(colored(value.get("stderr", ""), "red"))
                        print(colored(value.get("error", ""), "red"))
                    else:
                        print(colored("-" * 80, "green"))
                        print(colored(key, "green"))
                        print(colored(value.get("stdout", ""), "white"))
                        if value.get("stderr"):
                            print(colored(value.get("stderr", ""), "red"))
        except KeyboardInterrupt:
            print(colored("-" * 80, "red"))
            print(colored("Interrupted by user", "red"))

    def do_exec(self, line, prefix=None):
        """
        Executes a line in shell on specified boxes

        Syntax:
            exec <subset> "command -to 'execute'"
        """
        cmd = Command(line)
        nodes = self.find_nodes(cmd.get(0))
        command = " ".join(cmd.args[1:])
        if prefix is not None:
            command = prefix + " " + command
        self.execute(command, nodes)

    def do_sudo(self, line, postfix=None):
        """
        Executes a line in shell on specified boxes with sudo

        Syntax:
            sudo <subset> "command -to 'execute'"
        """
        prefix = "sudo"
        if postfix is not None:
            prefix = prefix + " " + postfix
        self.do_exec(line, prefix=prefix)

    def do_kubectl(self, line):
        """
        Executes a command on kubectl

        Syntax:
            kubectl <subset> "command -to 'execute'"
        """
        self.do_exec(line, prefix="kubectl")

    def do_etcdctl(self, line):
        """
        Uses etcdctl to query and control etcd on a remote machine

        Syntax:
            etcdctl <subset> "command -to 'execute'"
        """
        self.do_exec(line, prefix="etcdctl")

    def do_docker(self, line):
        """
        Executes sudo docker [args, like -a]

        Syntax:
            docker <subset> [args, like -a]
        """
        self.do_sudo(line, postfix="docker")


    ###########################################################################
    # KUBERNETES RELATED FUNCTIONALITY
    ###########################################################################

    def do_namespaces(self, line):
        """
            Prints all the namespaces available
        """
        for namespace in self.k8s_inventory.find_namespaces():
            print(namespace)

    def complete_deployments(self, text, line, begidx, endidx):
        """
            Auto-complete for k8s deployments
        """
        return filter_text_insensitive(self.k8s_inventory.find_namespaces(), text)

    def do_deployments(self, line):
        """
        List deployments
        Syntax:
            deployments [namespace=default]
        """
        cmd = Command(line)
        namespace = cmd.get(0)
        for deploy in self.k8s_inventory.find_deployments(
            namespace=namespace,
        ):
            print(deploy)

    def complete_pods(self, text, line, begidx, endidx):
        """
            Auto-complete for k8s pods
        """
        cmd = Command(line)
        namespace = cmd.get(1)
        if len(cmd) == 1 or (len(cmd) == 2 and not cmd.finished):
            return filter_text_insensitive(self.k8s_inventory.find_namespaces(), namespace)
        return []

    def do_pods(self, line):
        """
        List pods
        Syntax:
            pods namespace [selector]
        Selector is in the kubernetes native form: app=something,ver=1
        """
        cmd = Command(line)
        namespace = cmd.get(0)
        selector = cmd.get(1)
        for pod in self.k8s_inventory.find_pods(
            namespace=namespace,
            selector=selector
        ):
            print(colour_output(str(pod)))

    def complete_pods_for_deployment(self, text, line, begidx, endidx):
        """
            Auto-complete for k8s pods for deployment
        """
        cmd = Command(line)
        namespace = cmd.get(1)
        if len(cmd) == 1 or (len(cmd) == 2 and not cmd.finished):
            return filter_text_insensitive(self.k8s_inventory.find_namespaces(), namespace)
        else:
            op = cmd.get(2)
            return filter_text_insensitive(self.k8s_inventory.find_deployments(namespace), op)

    def do_pods_for_deployment(self, line):
        """
        List pods for a deployment
        Syntax:
            pods_for_deployment namespace deployment-name
        """
        cmd = Command(line)
        namespace = cmd.get(0)
        deployment_name = cmd.get(1)
        for pod in self.k8s_inventory.find_pods(
            namespace=namespace,
            deployment_name=deployment_name,
        ):
            print(colour_output(str(pod)))

    def do_cached_pods(self, line):
        """
        List the last set of pods loaded.
        Useful, when you want to kill a pod
        Syntax:
            cached_pods
        """
        pods = self.k8s_inventory.last_pods
        if not pods:
            return print("No pods loaded. Use `pods` to load a set to choose from")
        for pod in pods:
            print(colour_output(str(pod)))

    def complete_kill(self, text, line, begidx, endidx):
        """
            Auto-complete for k8s pods killing
        """
        # TODO
        return []

    def do_kill(self, line):
        """
        Kill a pod
        Syntax:
            kill pod-number
        To show the pod numbers, run cached_pods.
        Executes `sudo docker kill <container ID> for each container in the pod`
        """
        cmd = Command(line)
        pod_num = cmd.get(0)
        if pod_num is None:
            return
        try:
            pod_num = int(pod_num)
        except ValueError:
            return

        # find the pod
        pod = None
        for p in self.k8s_inventory.last_pods:
            if p.num == pod_num:
                pod = p
                break
        if pod is None:
            return print("Pod number not found.")
        if pod.state.lower() != 'running':
            return print("Can't kill pod with state", pod.state)

        # check with the user
        ans = False
        while ans not in ("y", "n"):
            print("Will delete pod '%s' through kubernetes API. Continue ? [y/n]: " % pod)
            ans = input().lower()
        if ans != "y":
            return print("Cancelling")

        # kill the pod
        success = self.executor.kill_pod(pod, self.inventory)
        if success:
            print("Done")
        else:
            print(colored("Could not delete the pod", "red"))




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from mock import MagicMock

from powerfulseal.node import Node, NodeInventory



@pytest.fixture
def nodes():
    return [
        Node(id="id1", ip="198.168.1.1", extIp="10.22.66.1", az="AZ1", no=1, name="node1"),
        Node(id="id2", ip="198.168.2.1", extIp="10.22.66.2", az="AZ2", no=2, name="node2"),
        Node(id="id3", ip="198.168.3.1", extIp="10.22.66.3", az="AZ2", no=3, name="node3"),
    ]

@pytest.fixture
def mock_driver(nodes):
    mock = MagicMock()
    mock.nodes = nodes
    def get_by_ip(ip):
        for node in mock.nodes:
            if (node.ip == ip) or (node.extIp == ip):
                return node
    mock.get_by_ip = get_by_ip
    return mock


def test_sync(nodes, mock_driver):
    inventory = NodeInventory(
        driver=mock_driver,
        restrict_to_groups={
            "TEST1": ["198.168.1.1"],
            "TEST2": ["198.168.1.1", "198.168.2.1"],
        }
    )
    inventory.sync()
    assert inventory.groups == {
        "TEST1": nodes[0:1],
        "TEST2": nodes[0:2]
    }
    assert inventory.get_azs() == ["AZ1", "AZ2"]

@pytest.mark.parametrize("query, expected_indices", [
    (None, [0, 1]),
    ("all", [0, 1]),
    ("id1,id2", [0, 1]),
    ("id1", [0]),
    ("id2", [1]),
    ("198.168.1.1", [0]),
    ("10.22.66.2", [1]),
    ("10.22.66.4", []),
    ("AZ2", [1]),
    ("2", [1]),
    ("node2", [1]),
    ("TEST2", [0, 1]),
    ("up", []),
    ("unknown", [0, 1]),
    ("something-weird", []),
])
def test_find(nodes, mock_driver, query, expected_indices):
    inventory = NodeInventory(
        driver=mock_driver,
        restrict_to_groups={
            "TEST1": ["198.168.1.1"],
            "TEST2": ["198.168.1.1", "198.168.2.1"],
        }
    )
    inventory.sync()
    assert list(inventory.find_nodes(query)) == [nodes[x] for x in expected_indices]


@pytest.mark.parametrize("ip, should_find, index", [
    ("198.168.2.1", True, 1),
    ("doesn't exist", False, None),
])
def test_get_by_ip(nodes, mock_driver, ip, should_find, index):
    inventory = NodeInventory(
        driver=mock_driver,
        restrict_to_groups={
            "TEST1": ["198.168.1.1"],
            "TEST2": ["198.168.1.1", "198.168.2.1"],
        }
    )
    inventory.sync()
    if should_find:
        assert inventory.get_node_by_ip(ip) is nodes[index]
    else:
        assert inventory.get_node_by_ip(ip) == None

def test_groups(mock_driver):
    inventory = NodeInventory(
        driver=mock_driver,
        restrict_to_groups={
            "TEST1": ["198.168.1.1"],
            "TEST2": ["198.168.1.1", "198.168.2.1"],
        }
    )
    inventory.sync()
    assert inventory.get_groups() == ["TEST1", "TEST2"]





# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from powerfulseal.node import Node, NodeState

EXAMPLE_NODE_ARGS = dict(
    id="someid",
    name="random_node",
    ip="198.198.1.1",
    extIp="31.41.51.61",
    az="az1",
    groups=["group1","group2"],
    state=NodeState.UP,
)

@pytest.fixture
def node():
    return Node(**EXAMPLE_NODE_ARGS)


def test_node_passthrough(node):
    for key, val in EXAMPLE_NODE_ARGS.items():
        assert key in node.__dict__
        assert val == node.__dict__.get(key)

def test_node_str(node):
    rep = str(node)
    for key, val in EXAMPLE_NODE_ARGS.items():
        assert key in rep
        assert str(val) in rep

def test_node_repr(node):
    rep = repr(node)
    for key, val in EXAMPLE_NODE_ARGS.items():
        assert key in rep
        assert str(val) in rep

def test_node_state(node):
    node.state = NodeState.UP
    assert str(node.state) == "UP"
    node.state = NodeState.DOWN
    assert str(node.state) == "DOWN"

@pytest.mark.parametrize("state", [
    "UP",
    "up",
    "something unexpected",
])
def test_node_raises_on_bad_state(state):
    with pytest.raises(ValueError):
        Node(id="something", state=state)

def test_nodes_are_deduplicated():
    collection = set()
    collection.add(Node(**EXAMPLE_NODE_ARGS))
    collection.add(Node(**EXAMPLE_NODE_ARGS))
    assert len(collection) == 1




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pkg_resources
import pytest
from powerfulseal.node.inventory import read_inventory_file_to_dict


def test_reads_sample_inventory():
    filename = pkg_resources.resource_filename('tests.node', 'example_inventory')
    groups = read_inventory_file_to_dict(filename)
    assert groups["groupa"] == ["192.168.1.1"]
    assert groups["groupb"] == ["192.168.2.1"]
    assert groups["groupc"] == ["192.168.3.1"]
    assert groups["workers:children"] == list(set(["192.168.1.1","192.168.2.1","192.168.3.1"]))



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from datetime import datetime

import pytest
from mock import MagicMock

from powerfulseal.execute import SSHExecutor, KubernetesExecutor
from powerfulseal.node import NodeInventory, Node
from powerfulseal.policy.label_runner import LabelRunner
from powerfulseal.k8s.pod import Pod

def test_kill_pod_APIcalling():

    k8s_client_mock = MagicMock()
    label_runner = LabelRunner(NodeInventory(None), None, None, KubernetesExecutor(k8s_client_mock))

    # Patch action of getting nodes to execute kill command on
    test_node = Node(1)
    get_node_by_ip_mock = MagicMock(return_value=test_node)
    label_runner.inventory.get_node_by_ip = get_node_by_ip_mock

    # Patch k8s inventory
    k8s_inventory = MagicMock()
    label_runner.k8s_inventory = k8s_inventory

    # Patch action of choosing container
    label_runner.k8s_inventory.k8s_client = k8s_client_mock

    delete_pods_mock = MagicMock()
    label_runner.k8s_inventory.k8s_client.delete_pods = delete_pods_mock
    
    metric_collector = MagicMock()
    label_runner.metric_collector = metric_collector

    add_pod_killed_metric_mock = MagicMock()
    label_runner.metric_collector.add_pod_killed_metric = add_pod_killed_metric_mock

    mock_pod = Pod(name='test', namespace='test')
    mock_pod.container_ids = ["docker://container1"]
    label_runner.kill_pod(mock_pod)

    delete_pods_mock.assert_called_with([mock_pod])
    add_pod_killed_metric_mock.assert_called_with(mock_pod)


def test_kill_pod_SSHing():    
    label_runner = LabelRunner(NodeInventory(None), None, None, SSHExecutor())

    # patch metrics collector
    label_runner.metric_collector = MagicMock()

    # Patch action of switching to SSHing mode
    k8s_inventory = MagicMock()
    label_runner.k8s_inventory = k8s_inventory
    
    # Patch action of getting nodes to execute kill command on
    test_node = Node(1)
    get_node_by_ip_mock = MagicMock(return_value=test_node)
    label_runner.inventory.get_node_by_ip = get_node_by_ip_mock

    # Patch action of choosing container
    execute_mock = MagicMock()
    label_runner.executor.execute = execute_mock

    mock_pod = Pod(name='test', namespace='test')
    mock_pod.container_ids = ["docker://container1"]
    mock_pod.labels = {"seal/force-kill": "false"}
    label_runner.kill_pod(mock_pod)

    mock_pod = Pod(name='test', namespace='test')
    mock_pod.container_ids = ["docker://container1"]
    mock_pod.labels = {}
    label_runner.kill_pod(mock_pod)

    execute_mock.assert_called_with("sudo docker kill -s SIGTERM container1", nodes=[test_node])


def test_kill_pod_forced_SSHing():
    label_runner = LabelRunner(NodeInventory(None), None, None, SSHExecutor())

    # patch metrics collector
    label_runner.metric_collector = MagicMock()

    # Patch action of switching to SSHing mode
    k8s_inventory = MagicMock()
    label_runner.k8s_inventory = k8s_inventory
    
    # Patch action of getting nodes to execute kill command on
    test_node = Node(1)
    get_node_by_ip_mock = MagicMock(return_value=test_node)
    label_runner.inventory.get_node_by_ip = get_node_by_ip_mock

    # Patch action of choosing container
    execute_mock = MagicMock()
    label_runner.executor.execute = execute_mock

    mock_pod = Pod(name='test', namespace='test')
    mock_pod.container_ids = ["docker://container1"]
    mock_pod.labels = {"seal/force-kill": "true"}
    label_runner.kill_pod(mock_pod)
    execute_mock.assert_called_once_with("sudo docker kill -s SIGKILL container1", nodes=[test_node])


def test_filter_is_enabled():
    label_runner = LabelRunner(None, None, None, None)

    pods = [
        Pod(name='test', namespace='test', labels={'seal/enabled': 'true'}),
        Pod(name='test', namespace='test', labels={'seal/enabled': 'false'}),
        Pod(name='test', namespace='test', labels={'seal/enabled': 'asdf'}),
        Pod(name='test', namespace='test', labels={'bla': 'bla'}),
        Pod(name='test', namespace='test', annotations={'seal/enabled': 'true'}),
        Pod(name='test', namespace='test', annotations={'seal/enabled': 'false'}),
        Pod(name='test', namespace='test', annotations={'seal/enabled': 'true'}, labels={'seal/enabled': 'false'}),
        Pod(name='test', namespace='test', annotations={'seal/enabled': 'false'}, labels={'seal/enabled': 'true'})
    ]

    filtered_pods = label_runner.filter_is_enabled(pods)

    assert len(filtered_pods) is 3
    assert filtered_pods[0] is pods[0]
    assert filtered_pods[1] is pods[4]
    assert filtered_pods[2] is pods[7]

@pytest.mark.parametrize("proba", [
    0.3,
    0.1,
    0.5,
    0.9
])
def test_filter_kill_probability(proba):
    random.seed(7)  # make the tests deterministic
    SAMPLES = 100000

    label_runner = LabelRunner(None, None, None, None)
    pod = Pod(name='test', namespace='test')
    pod.labels = {'seal/kill-probability': str(proba)}

    agg_len = 0.0
    for _ in range(SAMPLES):
        agg_len += len(label_runner.filter_kill_probability([pod]))
    assert float(agg_len) / SAMPLES == pytest.approx(proba, 0.01)


def test_filter_day_time():
    label_runner = LabelRunner(None, None, None, None)

    pod = Pod(name="test", namespace="test")
    pod.labels = {
        "seal/start-time": "10-00-00",
        "seal/end-time": "17-30-00",
        "seal/days": "mon,tue,wed,thu,fri,sat,sun"
    }

    for day in range(7):
        now = datetime.now()
        # check that it works for all the days of the week, as specified above
        now.replace(day=day+1)
        # set the microseconds to 0, to pass the inclusive start
        now.replace(microsecond=0)
        test_cases = [
            (now.replace(hour=8, minute=0, second=0), False, "far too early"),
            (now.replace(hour=9, minute=59, second=59), False, "just too early"),
            (now.replace(hour=10, minute=0, second=0), True, "inclusive start"),
            (now.replace(hour=13, minute=0, second=0), True, "within"),
            (now.replace(hour=17, minute=30, second=0), False, "exclusive end"),
            (now.replace(hour=17, minute=30, second=1), False, "just too late"),
            (now.replace(hour=20, minute=0, second=0), False, "far too late")
        ]

        for test_case in test_cases:
            assert (len(label_runner.filter_day_time([pod], test_case[0])) == 1) == test_case[1]


def test_get_integer_days_from_days_label():
    label_runner = LabelRunner(None, None, None, None)
    integer_days = label_runner.get_integer_days_from_days_label("mon,abc,tue,wed,thu,thur,fri,sat,sun,bla")
    assert integer_days == [0, 1, 2, 3, 4, 5, 6]


def test_process_time_label():
    label_runner = LabelRunner(None, None, None, None)

    with pytest.raises(ValueError) as _:
        label_runner.process_time_label("10-00-0")

    with pytest.raises(ValueError) as _:
        label_runner.process_time_label("-1-00-00")

    with pytest.raises(ValueError) as _:
        label_runner.process_time_label("24-00-00")

    assert (1, 22, 21) == label_runner.process_time_label("01-22-21")




# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import pkg_resources
from mock import MagicMock
import mock

from powerfulseal.policy import PolicyRunner


def test_default_policy_validates():
    assert PolicyRunner.is_policy_valid(PolicyRunner.DEFAULT_POLICY)


def test_example_config_validates():
    filename = pkg_resources.resource_filename("tests.policy", "example_config.yml")
    policy = PolicyRunner.load_file(filename)
    assert PolicyRunner.is_policy_valid(policy)

crd_scenario_name = "test"

def get_scenarios_mock():
    return [{
        "name": crd_scenario_name,
        "steps": [
            {"podAction": {
                "matches": [{"labels": {"namespace": "powerfulseal-sandbox", "selector": "app=nginx"}}],
                "actions": [{"kill": {"probability": 1}}]
            }}
        ]
    }]

def test_parses_config_correctly(monkeypatch):
    with mock.patch('powerfulseal.k8s.k8s_client') as k8s_client:
        k8s_client.get_scenarios = get_scenarios_mock
        sleep_mock = MagicMock()
        monkeypatch.setattr("time.sleep", sleep_mock)
        filename = pkg_resources.resource_filename(
            "tests.policy", "example_config2.yml")
        runner = PolicyRunner(filename, k8s_client)
        policy = runner.read_policy()
        inventory = MagicMock()
        k8s_inventory = MagicMock()
        driver = MagicMock()
        executor = MagicMock()
        LOOPS = policy.get("config").get("runStrategy").get("runs")
        assert policy.get("scenarios")[1].get("name") == crd_scenario_name
        runner.run(inventory, k8s_inventory, driver, executor)
        assert sleep_mock.call_count == LOOPS
        for call in sleep_mock.call_args_list:
            args, _ = call
            assert 77 <= args[0] <= 78



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import time
import sys
import copy

import jsonschema
import yaml
import pkgutil
from powerfulseal import makeLogger
from .scenario import Scenario

logger = makeLogger(__name__)


class PolicyRunner():
    """ Reads, validates and executes a JSON schema-compliant policy
    """
    DEFAULT_POLICY = {
        "scenarios": []
    }

    def __init__(self, config_file, k8s_client, logger=None):
        self.config_file = config_file
        self.k8s_client = k8s_client
        self.logger = logger or makeLogger(__name__)

    def read_policy(self):
        """
            Read configuration
        """
        if self.config_file is None:
            policy = copy.deepcopy(PolicyRunner.DEFAULT_POLICY)
        else:
            policy = PolicyRunner.load_file(self.config_file)

        # Load scenarios from K8S crd extending file scenarios
        scenarios = self.k8s_client.get_scenarios()
        policy['scenarios'].extend(scenarios)
        if not PolicyRunner.is_policy_valid(policy):
            self.logger.error("Policy not valid. See log output above.")
            return sys.exit(1)
        return policy

    @classmethod
    def get_schema(cls):
        """ Reads the schema from the file
        """
        data = pkgutil.get_data(__name__, "ps-schema.yaml")
        return yaml.safe_load(data)

    @classmethod
    def load_file(cls, filename):
        with open(filename, "r") as f:
            return yaml.safe_load(f.read())

    @classmethod
    def is_policy_valid(cls, policy, schema=None):
        schema = schema or cls.get_schema()
        try:
            jsonschema.validate(policy, schema)
        except jsonschema.ValidationError as error:
            logger.error(policy)
            logger.error(error)
            return False
        return True

    def run(self, inventory, k8s_inventory, driver, executor,
            metric_collector=None):
        """ Runs a policy forever
        """
        policy = self.read_policy()
        loops = policy.get("config", {}).get("runStrategy", {}).get("runs", None)
        while loops is None or loops > 0:
            policy = self.read_policy()
            config = policy.get("config", {}).get("runStrategy", {})
            should_randomize = config.get("strategy") == "random"
            wait_min = config.get("minSecondsBetweenRuns", 0)
            wait_max = config.get("maxSecondsBetweenRuns", 300)
            exitConfig = policy.get("config", {}).get("exitStrategy", {})
            exitStrategy = exitConfig.get("strategy", "report")
            scenarios = [
                Scenario(
                    name=item.get("name"),
                    schema=item,
                    inventory=inventory,
                    k8s_inventory=k8s_inventory,
                    driver=driver,
                    executor=executor,
                    metric_collector=metric_collector
                )
                for item in policy.get("scenarios", [])
            ]

            if should_randomize:
                random.shuffle(scenarios)
            for scenario in scenarios:
                ret = scenario.execute()
                if not ret:
                    if exitStrategy == "fail-fast":
                        logger.error("Exiting early")
                        return False
                    else:
                        logger.error("Scenario failed, reporting and carrying on")
            sleep_time = int(random.uniform(wait_min, wait_max))
            logger.info("Sleeping for %s seconds", sleep_time)
            time.sleep(sleep_time)
            if loops is not None:
                loops -= 1
        logger.info("All done here!")
        return True



# Copyright 2017 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import yaml
from configargparse import ArgumentParser, YAMLConfigFileParser
import logging
import coloredlogs
import textwrap
import sys
import os

from powerfulseal import makeLogger
import powerfulseal.version
from powerfulseal.k8s.metrics_server_client import MetricsServerClient
from prometheus_client import start_http_server
from powerfulseal.metriccollectors import StdoutCollector, PrometheusCollector, DatadogCollector
from powerfulseal.policy.label_runner import LabelRunner
from powerfulseal.web.server import start_server, ServerStateLogHandler
from ..node import NodeInventory
from ..node.inventory import read_inventory_file_to_dict
from ..clouddrivers import OpenStackDriver, AWSDriver, NoCloudDriver, AzureDriver, GCPDriver
from ..execute import SSHExecutor, KubernetesExecutor
from ..k8s import K8sClient, K8sInventory
from .pscmd import PSCmd
from ..policy import PolicyRunner

KUBECONFIG_DEFAULT_PATH = "~/.kube/config"

def parse_kubeconfig(args):
    """
        if explicitly set, use the --kubeconfig value
        otherwise, check if KUBECONFIG is set
        if not, check if there is `~/.kube/config` available
        else try to build in-cluster config
    """
    logger = makeLogger(__name__)
    kube_config = None
    expanded_home_kube_config_path = os.path.expanduser(KUBECONFIG_DEFAULT_PATH)
    if args.kubeconfig:
        kube_config = args.kubeconfig
        logger.info("Creating kubernetes client with config %s from --kubeconfig flag", kube_config)
    elif os.environ.get("KUBECONFIG"):
        kube_config = os.path.expanduser(os.environ.get("KUBECONFIG"))
        logger.info("Creating kubernetes client with config %s from KUBECONFIG env var", kube_config)
    elif os.path.exists(expanded_home_kube_config_path):
        kube_config = expanded_home_kube_config_path
        logger.info("Creating kubernetes client with config %s (path found for backwards compatibility)", kube_config)
    else:
        logger.info("Creating kubernetes client with in-cluster config")
    return kube_config

def add_kubernetes_options(parser):
    # Kubernetes
    args_kubernetes = parser.add_argument_group('Kubernetes settings')
    args_kubernetes.add_argument(
        '--kubeconfig',
        help='Location of kube-config file',
        type=os.path.expanduser
    )
    args_kubernetes.add_argument(
        '--use-pod-delete-instead-of-ssh-kill',
        help='DEPRECATED! If set, will not require SSH (will delete pods instead) - DEPRECATED - now triggers --mode=kubernetes',
        default=False,
        action='store_true',
    )
    args_kubernetes.add_argument(
        '--execution-mode',
        help=(
            'PowerfulSeal supports two ways of injecting failure: '
            '1) through SSH and 2) by scheduling containers in Kubernetes. '
            'Use of SSH leverages Docker directly and removes Kubernetes from the equation. It\'s typically faster too. '
            'But it requires SSH access to all nodes in the cluster. '
            'Alternatively, we can rely on Kubernetes to schedule our chaos pods. Slower, less reliable, but requires no special setup. '
            'The default is now to use Kubernetes'
        ),
        default="kubernetes",
        choices=["kubernetes", "ssh"]
    )

def add_ssh_options(parser):
    # SSH
    args_ssh = parser.add_argument_group('SSH settings')
    args_ssh.add_argument(
        '--remote-user',
        default=os.environ.get("PS_REMOTE_USER", "cloud-user"),
        help="the of the user for the ssh connections",
    )
    args_ssh.add_argument(
        '--ssh-allow-missing-host-keys',
        default=False,
        action='store_true',
        help='Allow connection to hosts not present in known_hosts',
    )
    args_ssh.add_argument(
        '--override-ssh-host',
        help=(
            'If you\'d like to execute all commands on a different host '
            '(for example for minikube) you can override it here'
        )
    )
    args_ssh.add_argument(
        '--use-private-ip',
        default=False,
        action='store_true',
        help='Use the private IP of each node (vs public IP)',
    )
    ssh_options = args_ssh.add_mutually_exclusive_group()
    ssh_options.add_argument(
        '--ssh-path-to-private-key',
        default=os.environ.get("PS_PRIVATE_KEY"),
        help='Path to ssh private key',
    )
    ssh_options.add_argument(
        '--ssh-password',
        default=os.environ.get("PS_SSH_PASSWORD"),
        help='ssh password',
    )
    args_ssh.add_argument(
        '--ssh-kill-command',
        default=os.environ.get("PS_SSH_KILL_COMMAND", "sudo docker kill -s {signal} {container_id}"),
        help='The command to execute on remote host to kill the {container_id}',
    )

def add_inventory_options(parser):
    # Inventory
    args = parser.add_argument_group('Inventory settings')
    inventory_options = args.add_mutually_exclusive_group(required=False)
    inventory_options.add_argument('-i', '--inventory-file',
        default=os.environ.get("INVENTORY_FILE"),
        help=('the inventory file (in ini format) of groups '
              'of hosts to work with')
    )
    inventory_options.add_argument('--inventory-kubernetes',
        help='reads all kubernetes cluster nodes as inventory',
        action='store_true',
    )

def add_cloud_options(parser):
    # Cloud Driver
    args = parser.add_argument_group('Cloud settings')
    cloud_options = args.add_mutually_exclusive_group()
    cloud_options.add_argument('--openstack',
        default=os.environ.get("OPENSTACK_CLOUD"),
        action='store_true',
        help="use OpenStack cloud provider",
    )
    cloud_options.add_argument('--aws',
        default=os.environ.get("AWS_CLOUD"),
        action='store_true',
        help="use AWS cloud provider",
    )
    cloud_options.add_argument('--azure',
        default=os.environ.get("AZURE_CLOUD"),
        action='store_true',
        help="use Azure cloud provider",
    )
    cloud_options.add_argument('--gcp',
        default=os.environ.get("GCP_CLOUD"),
        action='store_true',
        help="use GCP cloud provider",
    )
    cloud_options.add_argument('--no-cloud',
        default=os.environ.get("NO_CLOUD"),
        action='store_true',
        help="don't use cloud provider",
    )
    # other options
    args.add_argument('--openstack-cloud-name',
        default=os.environ.get("OPENSTACK_CLOUD_NAME"),
        help="optional name of the open stack cloud from your config file to use",
    )
    args.add_argument('--azure-resource-group-name',
        default=os.environ.get("AZURE_RESORUCE_GROUP_NAME"),
        help="optional name of the Azure vm cluster resource group. Used to determine azure-node-resource-group-name.",
    )
    args.add_argument('--azure-node-resource-group-name',
        default=os.environ.get("AZURE_NODE_RESORUCE_GROUP_NAME"),
        help="name of the Azure vm cluster node resource group",
    )
    args.add_argument('--gcp-config-file',
        default=os.environ.get("GCP_CONFIG_FILE"),
        help="name of the gcloud config file (in json) to use instead of the default one",
    )

def add_namespace_options(parser):
    args = parser.add_argument_group('Kubernetes options')
    args.add_argument('--kubernetes-namespace',
        default='default',
        help='Namespace to use for label mode '
             'or a comma-separated list of namespaces'
             '(set to blank for all namespaces)'
    )

def add_policy_options(parser):
    # Policy
    args = parser.add_argument_group('Policy settings')
    args.add_argument('--policy-file',
        default=os.environ.get("POLICY_FILE"),
        help='the policy file to run',
        required=False
    )

def add_run_options(parser):
    # policy settings
    run_args = parser.add_argument_group(
        title='Policy settings'
    )
    run_args.add_argument('--min-seconds-between-runs',
        help='Minimum number of seconds between runs',
        default=0,
        type=int
    )
    run_args.add_argument('--max-seconds-between-runs',
        help='Maximum number of seconds between runs',
        default=300,
        type=int
    )

def add_metrics_options(parser):
    # metrics settings
    autonomous_args = parser.add_argument_group(
        title='Metrics settings'
    )
    metric_options = autonomous_args.add_mutually_exclusive_group(required=False)
    metric_options.add_argument('--stdout-collector',
        default=os.environ.get("STDOUT_COLLECTOR"),
        action='store_true',
        help="print metrics collected to stdout"
    )
    metric_options.add_argument('--prometheus-collector',
        default=os.environ.get("PROMETHEUS_COLLECTOR"),
        action='store_true',
        help="store metrics in Prometheus and expose metrics over a HTTP server"
    )
    metric_options.add_argument('--datadog-collector',
        default=os.environ.get("DATADOG_COLLECTOR"),
        action='store_true',
        help="send collected metrics to Datadog using DogStatsD"
    )

    args_prometheus = parser.add_argument_group('Prometheus settings')
    args_prometheus.add_argument(
        '--prometheus-host',
        default='0.0.0.0',
        help=(
            'Host to expose Prometheus metrics via the HTTP server when using '
            'the --prometheus-collector flag'
        ),
    )
    args_prometheus.add_argument(
        '--prometheus-port',
        default=9000,
        help=(
            'Port to expose Prometheus metrics via the HTTP server '
            'when using the --prometheus-collector flag'
        ),
        type=check_valid_port
    )

def add_common_options(parser):
    add_kubernetes_options(parser)
    add_cloud_options(parser)
    add_inventory_options(parser)
    add_ssh_options(parser)


def check_valid_port(value):
    parsed = int(value)
    min_port = 0
    max_port = 65535
    if parsed < min_port or parsed > max_port:
        raise argparse.ArgumentTypeError("%s is an invalid port number" % value)
    return parsed


def parse_args(args):
    parser = ArgumentParser(
        config_file_parser_class=YAMLConfigFileParser,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        default_config_files=['~/.config/seal', '~/.seal'],
        description=textwrap.dedent("""\
            PowerfulSeal
            The Chaos Engineering tool for Kubernetes

        """),
    )

    # General settings
    parser.add_argument(
        '-c', '--config',
        is_config_file=True,
        env_var="CONFIG",
        help='Config file path',
    )
    parser.add_argument('-v', '--verbose',
        action='count',
        help='Verbose logging.'
    )
    parser.add_argument('-s', '--silent',
        action='count',
        help='Silent logging.'
    )
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='%(prog)s {version}'.format(version=powerfulseal.version.__version__),
        help='Version.',
    )
    # subparsers
    subparsers = parser.add_subparsers(
        title='MODES OF OPERATION',
        description=(
            'Pick one of the following options to start the Seal in the '
            'specified mode. Learn more at '
            'https://github.com/bloomberg/powerfulseal#introduction'
        ),
        dest='mode'
    )

    ##########################################################################
    # INTERACTIVE MODE
    ##########################################################################
    parser_interactive = subparsers.add_parser('interactive',
        help=(
            'Starts an interactive CLI, which allows to manually issue '
            'commands on pods and nodes and provides a sweet autocomplete. '
            'If you\'re reading this for the first time, you should probably '
            'start here. '
        ),
    )
    add_common_options(parser_interactive)


    ##########################################################################
    # AUTONOMOUS MODE
    ##########################################################################
    parser_autonomous = subparsers.add_parser('autonomous',
        help=(
            'This is the main mode of operation. It reads the policy file and executes it.'
        ),
    )
    add_common_options(parser_autonomous)
    add_policy_options(parser_autonomous)
    add_metrics_options(parser_autonomous)

    # web ui settings
    web_args = parser_autonomous.add_argument_group(
        title='Web UI settings'
    )
    web_args.add_argument(
        '--headless',
        help='Doesn\'t start the UI, just runs the policy',
        action='store_true'
    )
    web_args.add_argument(
        '--host',
        help='Specify host for the PowerfulSeal web server',
        default=os.environ.get('HOST', '0.0.0.0')
    )
    web_args.add_argument(
        '--port',
        help='Specify port for the PowerfulSeal web server',
        default=int(os.environ.get('PORT', '8000')),
        type=check_valid_port
    )
    web_args.add_argument(
        '--accept-proxy-headers',
        help='Set this flag for the webserver to accept X-Forwarded-* headers',
        action='store_true'
    )

    ##########################################################################
    # LABEL MODE
    ##########################################################################
    parser_label = subparsers.add_parser('label',
        help=(
            'Starts in label mode. '
            'It reads Kubernetes pods in a specified namespace, and checks '
            ' their \'seal/*\' labels to decide which ones to kill.'
            'There is no policy needed in this mode. '
            'To learn about supported labels, read more at '
            'https://github.com/bloomberg/powerfulseal/ '
        ),
    )
    add_common_options(parser_label)
    add_namespace_options(parser_label)
    add_run_options(parser_label)
    add_metrics_options(parser_label)

    ##########################################################################
    # VALIDATE POLICY MODE
    ##########################################################################
    parser_validate_policy = subparsers.add_parser('validate',
        help=(
            'Validates any file against the policy schema, returns.'
            'You can use this to check that your policy is correct, '
            'before using it in autonomous mode.'
        )
    )
    add_policy_options(parser_validate_policy)

    return parser.parse_args(args=args)


def main(argv):
    """
        The main function to invoke the powerfulseal cli
    """
    args = parse_args(args=argv)

    if args.mode is None:
        return parse_args(['--help'])

    ##########################################################################
    # VALIDATE POLICY MODE
    ##########################################################################
    if args.mode == 'validate':
        policy = PolicyRunner.load_file(args.policy_file)
        if PolicyRunner.is_policy_valid(policy):
            return print('OK')
        print("Policy not valid. See log output above.")
        return sys.exit(1)

    ##########################################################################
    # LOGGING
    ##########################################################################
    # this is to calm down the flask stdout
    # calm down the workzeug
    logging.getLogger("werkzeug").setLevel(logging.ERROR)
    try:
        import click
        def echo(*args, **kwargs):
            pass
        click.echo = echo
        click.secho = echo
    except:
        pass

    # parse the verbosity flags
    if args.silent == 1:
        log_level = logging.WARNING
    elif args.silent == 2:
        log_level = logging.ERROR
    elif not args.verbose:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    server_log_handler = ServerStateLogHandler()
    server_log_handler.setLevel(log_level)

    # do a basic config with the server log handler
    logging.basicConfig(level=log_level, handlers=[server_log_handler])
    # this installs a stdout handler by default to the root
    coloredlogs.install(
        level=log_level,
        fmt='%(asctime)s %(levelname)s %(name)s %(message)s'
    )

    # the main cli handler
    logger = makeLogger(__name__)
    logger.setLevel(log_level)
    logger.info("verbosity: %s; log level: %s; handler level: %s", args.verbose, logging.getLevelName(logger.getEffectiveLevel()), logging.getLevelName(log_level) )

    ##########################################################################
    # KUBERNETES
    ##########################################################################
    kube_config = parse_kubeconfig(args)
    k8s_client = K8sClient(kube_config=kube_config)
    operation_mode = args.execution_mode
    # backwards compatibility
    if args.use_pod_delete_instead_of_ssh_kill:
        operation_mode = "kubernetes"
    k8s_inventory = K8sInventory(
        k8s_client=k8s_client,
    )

    ##########################################################################
    # CLOUD DRIVER
    ##########################################################################
    if args.openstack:
        logger.info("Building OpenStack driver")
        driver = OpenStackDriver(
            cloud=args.openstack_cloud_name,
        )
    elif args.aws:
        logger.info("Building AWS driver")
        driver = AWSDriver()
    elif args.azure:
        logger.info("Building Azure driver")
        driver = AzureDriver(
            cluster_rg_name=args.azure_resource_group_name,
            cluster_node_rg_name=args.azure_node_resource_group_name,
        )
    elif args.gcp:
        logger.info("Building GCP driver")
        driver = GCPDriver(config=args.gcp_config_file)
    else:
        logger.info("No cloud driver - some functionality disabled")
        driver = NoCloudDriver()

    ##########################################################################
    # INVENTORY
    ##########################################################################
    if args.inventory_file:
        logger.info("Reading inventory from %s", args.inventory_file)
        groups_to_restrict_to = read_inventory_file_to_dict(
            args.inventory_file
        )
    else:
        logger.debug("Attempting to read the inventory from kubernetes")
        groups_to_restrict_to = k8s_client.get_nodes_groups()

    logger.debug("Restricting inventory to %s" % groups_to_restrict_to)

    inventory = NodeInventory(
        driver=driver,
        restrict_to_groups=groups_to_restrict_to,
    )
    inventory.sync()

    ##########################################################################
    # SSH EXECUTOR
    ##########################################################################
    if operation_mode == "kubernetes":
        executor = KubernetesExecutor(
            k8s_client=k8s_client,
        )
    else:
        if args.use_private_ip:
            logger.info("Using each node's private IP address")
        if args.override_ssh_host:
            logger.info("Using each overriten host: %s", args.override_ssh_host)
        executor = SSHExecutor(
            user=args.remote_user,
            ssh_allow_missing_host_keys=args.ssh_allow_missing_host_keys,
            ssh_path_to_private_key=args.ssh_path_to_private_key,
            override_host=args.override_ssh_host,
            ssh_password=args.ssh_password,
            use_private_ip=args.use_private_ip,
            ssh_kill_command=args.ssh_kill_command,
        )

    ##########################################################################
    # INTERACTIVE MODE
    ##########################################################################
    if args.mode == 'interactive':
        # create a command parser
        cmd = PSCmd(
            inventory=inventory,
            driver=driver,
            executor=executor,
            k8s_inventory=k8s_inventory,
        )
        logger.info("STARTING INTERACTIVE MODE")
        while True:
            try:
                cmd.cmdloop()
            except GeneratorExit:
                print("Exiting")
                sys.exit(0)
            except KeyboardInterrupt:
                print()
                print("Ctrl-c again to quit")
            try:
                input()
            except KeyboardInterrupt:
                sys.exit(0)
        return

    ##########################################################################
    # METRICS
    ##########################################################################
    metric_collector = StdoutCollector()
    if args.prometheus_collector:
        flask_debug = os.environ.get("FLASK_DEBUG")
        flask_env = os.environ.get("FLASK_ENVIROMENT")
        if flask_debug is not None or (flask_env is not None and flask_env != "production"):
            logger.error("PROMETHEUS METRICS NOT SUPPORTED WHEN USING FLASK RELOAD. NOT STARTING THE SERVER")
        else:
            logger.info("Starting prometheus metrics server (%s:%s)", args.prometheus_host, args.prometheus_port)
            start_http_server(args.prometheus_port, args.prometheus_host)
            metric_collector = PrometheusCollector()
    elif args.datadog_collector:
        logger.info("Starting datadog collector")
        metric_collector = DatadogCollector()
    else:
        logger.info("Using stdout metrics collector")

    ##########################################################################
    # AUTONOMOUS MODE
    ##########################################################################
    if args.mode == 'autonomous':

        runner = PolicyRunner(args.policy_file, k8s_client, logger)

        # run the metrics server if requested
        if not args.headless:
            # start the server
            logger.info("Starting the UI server (%s:%s)", args.host, args.port)
            start_server(
                host=args.host,
                port=args.port,
                read_policy_fn=runner.read_policy,
                accept_proxy_headers=args.accept_proxy_headers,
                logger=server_log_handler,
            )
        else:
            logger.info("NOT starting the UI server")

        logger.info("STARTING AUTONOMOUS MODE")
        success = runner.run(
            inventory,
            k8s_inventory,
            driver,
            executor,
            metric_collector=metric_collector
        )
        if not success:
            logger.error("Policy runner finishes with an error")
            return sys.exit(1)
        return sys.exit(0)

    ##########################################################################
    # LABEL MODE
    ##########################################################################
    elif args.mode == 'label':
        label_runner = LabelRunner(
            inventory,
            k8s_inventory,
            driver,
            executor,
            min_seconds_between_runs=args.min_seconds_between_runs,
            max_seconds_between_runs=args.max_seconds_between_runs,
            namespace=args.kubernetes_namespace,
            metric_collector=metric_collector,
        )
        logger.info("STARTING LABEL MODE")
        label_runner.run()


def start():
    main(sys.argv[1:])


if __name__ == '__main__':
    start()



# Copyright 2018 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from powerfulseal import makeLogger
import random

import time
from datetime import datetime


class LabelRunner:
    """
    Matches labels of individual pods and kills them if appropriate.

    Labels supported:
    `seal/enabled`          either "true" or "false" (default: "false")
    `seal/force-kill`       either "true" or "false" (default: "false")
    `seal/kill-probability` a value between "0" and "1" inclusive (default: "1")
    `seal/days`             a comma-separated string consisting of "mon", "tue",
                            "wed", "thu", "fri", "sat", "sun" describing the
                            pod can be killed (default: "mon,tue,wed,thu,fri")
    `seal/start-time`       a value "HH-MM-SS" for the inclusive start boundary
                            of when a pod can be killed in the local timezone
                            (default: "10-00-00")
    `seal/end-time`         a value "HH-MM-SS" for the exclusive end boundary of
                            when a pod can be killed in the local timezone
                            (default: "17-30-00")
    """
    DEFAULT_DAYS_LABEL = "mon,tue,wed,thu,fri"
    DAY_STRING_TO_DATETIME = {'mon': 0, 'tue': 1, 'wed': 2, 'thu': 3, 'fri': 4,
                              'sat': 5, 'sun': 6}

    def __init__(self, inventory, k8s_inventory, driver, executor,
                 min_seconds_between_runs=0, max_seconds_between_runs=300, logger=None,
                 namespace=None, metric_collector=None):
        self.inventory = inventory
        self.k8s_inventory = k8s_inventory
        self.driver = driver
        self.executor = executor
        self.min_seconds_between_runs = min_seconds_between_runs
        self.max_seconds_between_runs = max_seconds_between_runs
        self.logger = logger or makeLogger(__name__)
        self.namespace = namespace
        self.metric_collector = metric_collector

    def run(self):
        while True:
            # Filter
            app_pods_in_namespace = self.k8s_inventory.find_pods(self.namespace)
            self.logger.info("Found %d pods" % len(app_pods_in_namespace))
            pods = self.filter_pods(app_pods_in_namespace)
            self.logger.info("Filtered to %d pods" % len(pods))

            # Execute
            for pod in pods:
                self.kill_pod(pod)

            # Sleep and sync
            sleep_time = int(random.uniform(self.min_seconds_between_runs, self.max_seconds_between_runs))
            self.logger.info("Sleeping for %s seconds", sleep_time)
            time.sleep(sleep_time)
            self.inventory.sync()

    def kill_pod(self, pod):
        # prep the arguments
        signal = "SIGTERM"
        if pod.get_label_or_annotation("seal/force-kill", "false") == "true":
            signal = "SIGKILL"
        # kill the pod
        success = None
        try:
            success = self.executor.kill_pod(pod, self.inventory, signal)
        except:
            success = False
            self.logger.exception("Exception while killing pod")
        # update the metrics
        if success:
            self.metric_collector.add_pod_killed_metric(pod)
            self.logger.info("Pod killed: %s", pod)
        else:
            self.metric_collector.add_pod_kill_failed_metric(pod)
            self.logger.error("Pod NOT killed: %s", pod)
        return success

    def filter_pods(self, pods):
        filters = [
            self.filter_is_enabled,
            self.filter_day_time,
            self.filter_kill_probability
        ]

        remaining_pods = pods
        for pod_filter in filters:
            remaining_pods = pod_filter(remaining_pods)

        return remaining_pods

    def filter_is_enabled(self, pods):
        return list(filter(lambda pod: pod.get_label_or_annotation("seal/enabled", "false") == "true", pods))

    def filter_kill_probability(self, pods):
        remaining_pods = []
        for pod in pods:
            # Retrieve probability value, performing validation
            try:
                probability = float(pod.get_label_or_annotation("seal/kill-probability", "1"))
                if probability < 0 or probability > 1:
                    raise ValueError
            except ValueError:
                self.logger.warning("Invalid float value - skipping pod %s/%s" % (pod.namespace, pod.name))
                continue

            p = random.random()
            if probability < p:
                continue

            remaining_pods.append(pod)

        return remaining_pods

    def filter_day_time(self, pods, now=None):
        remaining_pods = []
        now = now or datetime.now()

        for pod in pods:
            # Filter on days
            days_label = pod.get_label_or_annotation("seal/days", self.DEFAULT_DAYS_LABEL)
            if now.weekday() not in self.get_integer_days_from_days_label(days_label):
                continue

            # Filter on start time
            start_time_label = pod.get_label_or_annotation("seal/start-time", "10-00-00")
            try:
                hours, minutes, seconds = self.process_time_label(start_time_label)
                start_time = now.replace(hour=hours, minute=minutes, second=seconds, microsecond=0)
                if now < start_time:
                    continue
            except ValueError:
                self.logger("Invalid start time - skipping pod %s/%s" % (pod.namespace, pod.name))
                continue

            # Filter on end time
            end_time_label = pod.get_label_or_annotation("seal/end-time", "17-30-00")
            try:
                hours, minutes, seconds = self.process_time_label(end_time_label)
                end_time = now.replace(hour=hours, minute=minutes, second=seconds, microsecond=0)
                if now >= end_time:
                    continue
            except ValueError:
                self.logger("Invalid end time - skipping pod %s/%s" % (pod.namespace, pod.name))
                continue

            remaining_pods.append(pod)

        return remaining_pods

    def get_integer_days_from_days_label(self, label):
        """
        Returns a list of integer day of week values compatible with the datetime
        package representing the days which a pod can be killed
        """
        string_days = label.split(',')

        days = []
        for string_day in string_days:
            try:
                days.append(self.DAY_STRING_TO_DATETIME[string_day])
            except KeyError:
                self.logger.warning("Invalid day label encountered: %s" % string_day)

        return days

    def process_time_label(self, label):
        """
        Validates and return a 3-tuple of hour, minutes and seconds
        :raises: ValueError
        """

        # "HH:MM:SS" has eight characters
        if len(label) != 8:
            raise ValueError("Label has invalid length (must be in HH-MM-SS format")

        tokens = label.split('-')

        # Ensure tokens is a list of three values ('HH', 'MM', 'SS')
        if len(tokens) != 3 or not all(map(lambda x: len(x) == 2, tokens)):
            raise ValueError("Label be in HH-MM-SS format")

        hours = int(tokens[0])
        minutes = int(tokens[1])
        seconds = int(tokens[2])

        if hours < 0 or hours > 23 or minutes < 0 or minutes > 59 or seconds < 0 or seconds > 59:
            raise ValueError("Label must be in HH-MM-SS format")

        return hours, minutes, seconds
